import warnings
from importlib.util import find_spec

__all__ = ["main"]

if find_spec("click"):

    from torrent_models.cli.main import main
else:
    warnings.warn(
        "cli dependencies are not installed - install with torrent-models[cli]", stacklevel=2
    )



import sys
from abc import abstractmethod
from enum import StrEnum
from pathlib import Path
from typing import Annotated, NotRequired, TypeAlias

from annotated_types import Ge, Len
from pydantic import AfterValidator, AnyUrl, BaseModel, Field

from torrent_models.base import ConfiguredBase
from torrent_models.types.serdes import ByteStr

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class TorrentVersion(StrEnum):
    """Version of the bittorrent .torrent file spec"""

    v1 = "v1"
    v2 = "v2"
    hybrid = "hybrid"


FileName: TypeAlias = ByteStr
"""Placeholder in case specific validation is needed for filenames"""
FilePart: TypeAlias = ByteStr
"""Placeholder in case specific validation is needed for filenames"""
SHA1Hash = Annotated[bytes, Len(20, 20)]
SHA256Hash = Annotated[bytes, Len(32, 32)]


TrackerFields = TypedDict(
    "TrackerFields",
    {
        "announce": NotRequired[AnyUrl | str],
        "announce-list": NotRequired[list[list[AnyUrl]] | list[list[str]]],
    },
)
"""The `announce` and `announce-list`"""


def _is_abs(path: Path) -> Path:
    """File must be absolute. if it is relative and still exists, make absolute"""
    if path.exists() and not path.is_absolute():
        path = path.resolve()
    assert path.is_absolute(), "Path must be absolute"
    return path


def _is_rel(path: Path) -> Path:
    """File must be relative"""
    assert not path.is_absolute(), "Path must be relative"
    return path


AbsPath = Annotated[Path, AfterValidator(_is_abs)]
RelPath = Annotated[Path, AfterValidator(_is_rel)]


def _divisible_by_16kib(size: int) -> int:
    assert size >= (16 * (2**10)), "Size must be at least 16 KiB"
    assert size % (16 * (2**10)) == 0, "Size must be divisible by 16 KiB"
    return size


def _power_of_two(n: int) -> int:
    assert (n & (n - 1) == 0) and n != 0, "Piece size must be a power of two"
    return n


class GenericFileItem(ConfiguredBase):
    """
    File metadata object with file information from both v1 and v2 representations

    For use with :class:`.Torrent.files`
    """

    path: FileName
    length: Annotated[int, Ge(0)]
    attr: bytes | None = None
    pieces_root: bytes | None = Field(None, alias="pieces root")


class PieceRange(BaseModel):
    """
    Parent model for v1 and v2 piece ranges.

    Piece ranges provide some description of paths and byte ranges that correspond to a single
    verifiable piece and a method for verifying data against them.

    Since v1 and v2 data models are substantially different,
    their sub-models are also quite different, but provide a common interface through this ABC
    """

    piece_idx: int

    @abstractmethod
    def validate_data(self, data: list[bytes]) -> bool:
        """Check that the provided data matches the piece or root hash"""



"""
Types used only in model serialization and deserialization:
AKA types that do not represent concrete types/fields in a torrent file
"""

from datetime import datetime
from typing import Annotated, TypeVar, cast

from pydantic import AnyUrl, BaseModel, BeforeValidator, PlainSerializer, WrapSerializer
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler


def _timestamp_to_datetime(val: int | datetime) -> datetime:
    if isinstance(val, int | float):
        val = datetime.fromtimestamp(val)
    return val


def _datetime_to_timestamp(val: datetime) -> int:
    return round(val.timestamp())


UnixDatetime = Annotated[
    datetime, BeforeValidator(_timestamp_to_datetime), PlainSerializer(_datetime_to_timestamp)
]
EXCLUDE_STRINGIFY = ("piece_layers", "piece layers", "path")


def str_keys(
    value: dict | list | BaseModel, _isdict: bool | None = None
) -> dict | list | BaseModel:
    """
    Convert the byte-encoded keys of a bencoded dictionary to strings,
    avoiding value of keys we know to be binary encoded like piece layers.

    .. note: Performance Notes

        AKA why this function looks so weird.

        doing isinstance check at the start is slightly faster than EAFP and allocating new_value,
        but use th internal bool _isdict to avoid doing it twice when we have already checked
        the argument's type in a parent call

        since we only call this with fresh output from bencode, we know that we only have
        python builtins, and therefore we can check with the classname rather than isinstance,
        which is slightly faster. this means that this function is *not* a general purpose function,
        and should only be used on freshly decoded bencoded data!

        We iterate over lists in the ``dict`` leg rather than in the function root to avoid the
        final double type check for leaf nodes, checking only if they are dictionaries.
    """

    if _isdict or value.__class__.__name__ == "dict":
        new_value = {}
        value = cast(dict, value)
        for k, v in value.items():
            try:
                if isinstance(k, bytes):
                    k = k.decode("utf-8")
            except UnicodeDecodeError:
                # fine, e.g. for piece maps whose keys are not utf-8 encoded
                # any invalid keys will get caught in validation
                pass

            if k in EXCLUDE_STRINGIFY:
                new_value[k] = v
                continue

            if v.__class__.__name__ == "dict":
                v = str_keys(v, _isdict=True)
            elif v.__class__.__name__ == "list":
                v = [str_keys(item) for item in v]
            new_value[k] = v
        return new_value
    else:
        return value


def str_keys_list(value: list[dict | BaseModel]) -> list[dict | list | BaseModel]:
    return [str_keys(v) for v in value]


def _to_str(value: str | bytes) -> str:
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    return value


def _to_bytes(value: str | bytes, info: SerializationInfo) -> bytes | str:
    if info.context and info.context.get("mode") == "print":
        return str(value)

    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, AnyUrl):
        value = str(value).encode("utf-8")
    else:
        value = str(value).encode("utf-8")
    return value


ByteStr = Annotated[str, PlainSerializer(_to_bytes)]
ByteUrl = Annotated[AnyUrl, PlainSerializer(_to_bytes)]
_Inner = TypeVar("_Inner")


def _to_list(val: _Inner | list[_Inner]) -> list[_Inner]:
    if val and not isinstance(val, list):
        return [val]
    else:
        val = cast(list[_Inner], val)
        return val


def _from_list(
    val: list[_Inner] | None, handler: SerializerFunctionWrapHandler
) -> list[_Inner] | _Inner | None:
    partial = handler(val)
    if partial and len(partial) == 1:
        return partial[0]
    else:
        return partial


ListOrValue = Annotated[list[_Inner], BeforeValidator(_to_list), WrapSerializer(_from_list)]



"""
Convenience class used when creating a new torrent.

The :class:`.TorrentCreate` class provides some convenience fields
that allows some common fields to be declared in a version-agnostic way
at the top level, rather than nested within the infodict.
"""

import multiprocessing as mp
from pathlib import Path
from typing import Any, Self, cast

from pydantic import Field, model_validator

from torrent_models import Torrent, TorrentVersion
from torrent_models.compat import get_size
from torrent_models.const import DEFAULT_TORRENT_CREATOR, EXCLUDE_FILES
from torrent_models.hashing import HybridHasher, V1Hasher, add_padfiles
from torrent_models.hashing.v1 import sort_v1
from torrent_models.info import InfoDictHybrid, InfoDictHybridCreate, InfoDictV1, InfoDictV2
from torrent_models.torrent import TorrentBase
from torrent_models.types import (
    AbsPath,
    ByteStr,
    FileItem,
    TrackerFields,
    V1PieceLength,
    V2PieceLength,
)
from torrent_models.types.v2 import FileTree, PieceLayers


class TorrentCreate(TorrentBase):
    """
    A programmatically created torrent that may not have its hashes computed yet.

    Torrents may be created *either* by passing an info dict with all details,
    (with or without piece hashes), *or* by using a handful of convenience fields.
    E.g. rather than needing to pass a fully instantiated file tree,
    one can just pass a list of files to ``files``
    """

    _EXCLUDE = {
        "paths": True,
        "path_root": True,
        "trackers": True,
        "piece_length": True,
        "info": {"meta_version", "files", "file_tree", "piece_length"},
        "piece_layers": True,
    }
    """
    Exclude from model dumps when creating internal model dumps when generating.
    ie. because they are transformed by creation
    """

    # make parent types optional
    announce: ByteStr | None = None
    created_by: ByteStr | None = Field(DEFAULT_TORRENT_CREATOR, alias="created by")

    # convenience fields
    info: InfoDictHybridCreate = Field(default_factory=InfoDictHybridCreate)  # type: ignore
    paths: list[Path] | None = Field(
        None,
        description="""
        Convenience field for creating torrents from lists of files.
        Can be either relative or absolute.
        Paths must be located beneath the path root, passed either explicitly or using
        cwd (default).
        If absolute, paths are made relative to the path root.
        """,
    )
    path_root: AbsPath = Field(
        default_factory=Path, description="Path to interpret paths relative to"
    )

    trackers: list[ByteStr] | list[list[ByteStr]] | None = Field(
        None,
        description="Convenience method for declaring tracker lists."
        "If a flat list, put each tracker in a separate tier."
        "Otherwise, sublists indicate tiers.",
    )
    piece_length: V1PieceLength | V2PieceLength | None = Field(
        None, description="Convenience method for passing piece length"
    )
    similar: list[bytes] | None = Field(
        None, description="Infohashes of other torrents that might contain overlapping files"
    )

    @model_validator(mode="after")
    def no_duplicated_params(self) -> Self:
        """
        Ensure that values that can be set from the top level convenience fields aren't doubly set,

        We don't set the accompanying values in the infodict on instantiation because
        this object is intended to be a programmatic constructor object,
        so we expect these values to change and don't want to have to worry about
        state consistency in it -
        all values are gathered and validated when the torrent is generated.
        """
        if self.paths:
            assert not self.info.files, "Can't pass both paths and info.files"
            assert not self.info.file_tree, "Can't pass both paths and info.file_tree"
        if self.trackers:
            assert not self.announce, "Can't pass both trackers and announce"
            assert not self.announce_list, "Can't pass both trackers and announce_list"
        if self.piece_length:
            assert not self.info.piece_length, "Can't pass both piece_length and info.piece_length"
        return self

    @model_validator(mode="after")
    def name_from_path_root(self) -> Self:
        """If `name` is not provided, infer it from the path root"""
        if not self.info.name:
            self.info.name = self.path_root.name
        return self

    def generate(
        self, version: TorrentVersion | str, n_processes: int | None = 1, progress: bool = False
    ) -> Torrent:
        """
        Generate a torrent file, hashing its pieces and transforming convenience values
        to valid torrent values.
        """
        if isinstance(version, str):
            version = TorrentVersion.__members__[version]

        if n_processes is None:
            n_processes = mp.cpu_count()

        if version == TorrentVersion.v1:
            return self._generate_v1(n_processes, progress)
        elif version == TorrentVersion.v2:
            return self._generate_v2(n_processes, progress)
        elif version == TorrentVersion.hybrid:
            return self._generate_hybrid(n_processes, progress)
        else:
            raise ValueError(f"Unknown torrent version: {version}")

    def generate_libtorrent(
        self,
        version: TorrentVersion | str,
        output: Path | None = None,
        bencode: bool = False,
        progress: bool = False,
    ) -> dict | bytes:
        from torrent_models.libtorrent import create_from_model

        return create_from_model(
            self, version=version, progress=progress, output=output, bencode=bencode
        )

    def _generate_common(self) -> dict:
        # dump just the fields we want to have in the final torrent,
        # excluding top-level convenience fields (set in the generate methods),
        # and hash values which are created during generation
        dumped = self.model_dump(
            exclude_none=True,
            exclude=self._EXCLUDE,  # type: ignore
            by_alias=False,
        )

        dumped["info"]["piece_length"] = self._get_piece_length()
        if "similar" in dumped:
            dumped["info"]["similar"] = dumped["similar"]
            del dumped["similar"]
        dumped.update(self.get_trackers())
        return dumped

    def _generate_v1(self, n_processes: int, progress: bool = False, **kwargs: Any) -> Torrent:
        dumped = self._generate_common()

        paths = self.get_paths(clean=True, v1_order=True)
        file_items = self._get_v1_file_items(paths)

        if not self.info.files:
            dumped["info"]["files"] = file_items

        if "pieces" not in dumped["info"]:
            hasher = V1Hasher(
                paths=paths,
                piece_length=self._get_piece_length(),
                read_size=self._get_piece_length(),
                path_root=self.path_root,
                n_processes=n_processes,
                progress=progress,
                **kwargs,
            )
            hashes = hasher.process()
            hashes = [hash.hash for hash in sorted(hashes, key=lambda x: x.idx)]
            dumped["info"]["pieces"] = hashes
        info = InfoDictV1(**dumped["info"])
        del dumped["info"]
        return Torrent(info=info, **dumped)

    def _generate_v2(self, n_processes: int, progress: bool = False) -> Torrent:
        dumped = self._generate_common()
        paths = self.get_paths(clean=True, v1_order=False)

        if "piece_layers" not in dumped or "file_tree" not in dumped["info"]:
            piece_layers = PieceLayers.from_paths(
                paths=paths,
                piece_length=dumped["info"]["piece_length"],
                path_root=self.path_root,
                n_processes=n_processes,
                progress=progress,
            )
            dumped["piece_layers"] = piece_layers.piece_layers
            dumped["info"]["file_tree"] = piece_layers.file_tree.tree

        info = InfoDictV2(**dumped["info"])
        del dumped["info"]
        return Torrent(info=info, **dumped)

    def _generate_hybrid(self, n_processes: int, progress: bool = False) -> Torrent:
        dumped = self._generate_common()

        # Gather paths

        if (self.info.files or self.info.length) and self.info.file_tree:
            # check for inconsistent paths in v1 and v2 if both are present
            v1_paths = self._get_v1_paths()
            v1_items = self._get_v1_file_items(v1_paths)
            v2_paths = [Path(path) for path in FileTree.flatten_tree(self.info.file_tree)]
            if not len(v1_paths) == len(v2_paths) and not all(
                [v1p == v2p for v1p, v2p in zip(v1_paths, v2_paths)]
            ):
                raise ValueError(
                    "Both v1 files and v2 file tree present, but have inconsistent paths!"
                )
            paths = v2_paths
        else:
            paths = self.get_paths(clean=True, v1_order=False)
            # v1 files
            v1_items = self._get_v1_file_items(paths)

        # add padding to the v1 files
        v1_items = add_padfiles(v1_items, dumped["info"]["piece_length"])

        hasher = HybridHasher(
            paths=paths,
            path_root=self.path_root,
            piece_length=self.piece_length,
            read_size=self.piece_length,
            n_processes=n_processes,
            progress=progress,
        )
        hashes = hasher.process()
        piece_layers, v1_pieces = hasher.split_v1_v2(hashes)
        dumped["piece layers"] = piece_layers.piece_layers
        dumped["info"]["file tree"] = piece_layers.file_tree.tree
        dumped["info"]["pieces"] = v1_pieces
        if len(v1_items) == 1:
            dumped["info"]["name"] = v1_items[0].path[-1]
            dumped["info"]["length"] = v1_items[0].length
        else:
            dumped["info"]["files"] = v1_items

        info = InfoDictHybrid(**dumped["info"])
        del dumped["info"]
        return Torrent(info=info, **dumped)

    def get_paths(self, clean: bool = True, v1_order: bool = False) -> list[Path]:
        """
        Get paths specified in one of potentially several ways

        In order (first match is returned):
        - paths set in top level `paths` field
        - v2 file tree, if present
        - v1 `files`, if present
        - v1 `name`, if present with `length` set
        - iterate the files beneath the :attr:`.path_root`

        Args:
            clean (bool): clean and sort the files
            v1_order (bool): sort files in v1 order -
                first top-level files, then files in directories
                in case-sensitive alphanumeric order within those categories.
        """
        if self.paths:
            paths = self.paths.copy()
        elif self.info.file_tree is not None:
            tree = self.flat_files
            assert tree is not None
            paths = [Path(t) for t in tree]
        else:
            try:
                paths = self._get_v1_paths()
            except ValueError:
                # no V1 paths, get files beneath base-path
                paths = list(self.path_root.rglob("*"))

        if not paths:
            raise ValueError("No paths provided, and nothing found within path root!")

        if clean:
            paths = clean_files(paths, relative_to=self.path_root, v1=v1_order)
        return paths

    def _get_v1_paths(self, paths: list[Path] | None = None, v1_only: bool = False) -> list[Path]:
        if paths:
            files = paths
        elif self.paths:
            files = self.paths
        elif self.info.files:
            files = [Path(*f.path) for f in self.info.files]
        elif self.info.length and self.info.name is not None:
            files = [Path(self.info.name)]
        else:
            raise ValueError("paths not provided, and info.files and info.length are unset!")

        files = clean_files(files, relative_to=self.path_root, v1=v1_only)
        return files

    def _get_v1_file_items(self, paths: list[Path]) -> list[FileItem]:
        items = [FileItem(path=list(f.parts), length=get_size(self.path_root / f)) for f in paths]
        return items

    def get_trackers(
        self,
    ) -> TrackerFields:
        # FIXME: hideous
        if self.trackers:
            if isinstance(self.trackers[0], list):

                self.trackers = cast(list[list[str]], self.trackers)
                if len(self.trackers[0]) == 1 and len(self.trackers[0][0]) == 1:
                    return {"announce": self.trackers[0][0]}
                else:
                    return {"announce": self.trackers[0][0], "announce-list": self.trackers}
            else:
                self.trackers = cast(list[str], self.trackers)
                if len(self.trackers) == 1:
                    return {"announce": self.trackers[0]}
                else:
                    return {
                        "announce": self.trackers[0],
                        "announce-list": [[t] for t in self.trackers],
                    }
        else:
            trackers_: TrackerFields = {}
            if self.announce is not None:
                trackers_["announce"] = self.announce

            if self.announce_list is not None:
                trackers_["announce-list"] = self.announce_list

            return trackers_

    def _get_piece_length(self) -> int:
        piece_length = self.piece_length if self.piece_length else self.info.piece_length
        if piece_length is None:
            raise ValueError("No piece length provided!")
        return piece_length


def list_files(path: Path | str) -> list[Path]:
    """
    Recursively list files relative to path, sorting, excluding known system files
    """
    path = Path(path)
    if path.is_file():
        return [path]

    paths = list(path.rglob("*"))

    return clean_files(paths, path)


def clean_files(paths: list[Path], relative_to: Path, v1: bool = False) -> list[Path]:
    """
    Remove system files, and make paths relative to some directory root
    """
    cleaned = []
    for f in paths:
        if f.is_absolute():
            abs_f = f
            # no absolute paths in the torrent plz
            rel_f = f.relative_to(relative_to)
        else:
            abs_f = relative_to / f
            rel_f = f
        if not abs_f.exists():
            raise FileNotFoundError(
                f"File {abs_f} does not exist for path {f} relative to {relative_to}"
            )
        if abs_f.is_file() and f.name not in EXCLUDE_FILES:
            cleaned.append(rel_f)
    cleaned = sort_v1(cleaned) if v1 else sorted(cleaned, key=lambda f: f.as_posix())
    return cleaned



"""
Types used only in v1 (and hybrid) torrents
"""

import hashlib
from typing import Annotated, Self

from annotated_types import Ge
from pydantic import (
    AfterValidator,
    BeforeValidator,
    PlainSerializer,
    ValidationInfo,
    model_validator,
)
from pydantic_core.core_schema import SerializationInfo

from torrent_models.base import ConfiguredBase
from torrent_models.types.common import FilePart, PieceRange, SHA1Hash, _power_of_two

V1PieceLength = Annotated[int, AfterValidator(_power_of_two)]
"""
According to BEP 003: no specification, but "almost always a power of two",
so we validate that.
"""


def _validate_pieces(pieces: bytes | list[bytes]) -> list[bytes]:
    if isinstance(pieces, bytes):
        assert len(pieces) % 20 == 0, "Pieces length must be divisible by 20"
        pieces = [pieces[i : i + 20] for i in range(0, len(pieces), 20)]

    return pieces


def _serialize_pieces(
    pieces: list[bytes], info: SerializationInfo
) -> bytes | list[bytes] | list[str]:
    """Join piece lists to a big long byte string unless we're pretty printing"""
    if info.context and info.context.get("mode") == "print":
        ret = [p.hex() for p in pieces]
        if info.context.get("hash_truncate"):
            ret = [p[0:8] for p in ret]
        return ret
    return b"".join(pieces)


Pieces = Annotated[
    list[SHA1Hash], BeforeValidator(_validate_pieces), PlainSerializer(_serialize_pieces)
]


class FileItem(ConfiguredBase):
    length: Annotated[int, Ge(0)]
    path: list[FilePart]
    attr: bytes | None = None
    """
    BEP0047: A variable-length string. 
    When present the characters each represent a file attribute. 
    l = symlink, 
    x = executable, 
    h = hidden, 
    p = padding file. 
    Characters appear in no particular order and unknown characters should be ignored.
    """

    @property
    def is_padfile(self) -> bool:
        return self.attr is not None and b"p" in self.attr

    @model_validator(mode="after")
    def strict_padfile_naming(self, info: ValidationInfo) -> Self:
        """in strict mode, padfiles must be named `.pad/{length}"""
        if not self.is_padfile:
            return self
        if info.context and info.context.get("padding_path") == "strict":
            assert self.path == [
                ".pad",
                str(self.length),
            ], "strict mode - padfiles must be named `.pad/{length}`"
        return self


class FileItemRange(FileItem):
    """A File Item with a byte range, for use with V1PieceRange"""

    range_start: int
    range_end: int


class V1PieceRange(PieceRange):
    """
    Paths and byte ranges that correspond to a single v1
    """

    ranges: list[FileItemRange]
    piece_hash: SHA1Hash

    def validate_data(self, data: list[bytes]) -> bool:
        """
        Validate data against hash by concatenating bytes and comparing the SHA1 hash

        The user is responsible for providing all-zero bytestrings
        for any padding files in the indicated ranges
        """
        assert len(data) == len(
            self.ranges
        ), "Need to provide data chunks that correspond to each of the indicated file ranges"
        for range, d in zip(self.ranges, data):
            assert (range.range_end - range.range_start) == len(d), (
                "Provided data chunks must match the sizes indicated by the "
                "start and end ranges of each file range"
            )

        hasher = hashlib.new("sha1")
        for d in data:
            hasher.update(d)
        return self.piece_hash == hasher.digest()



import pytest
from pydantic_core._pydantic_core import ValidationError

from torrent_models.const import KiB
from torrent_models.info import InfoDictV1Base
from torrent_models.types import FileItem

no_padding = [
    FileItem(path=["no_padding"], length=14 * KiB),
    FileItem(path=["sup"], length=7 * KiB),
]
correct_padding = [
    FileItem(path=["correct"], length=16 * KiB),
    FileItem(path=["hey"], length=14 * KiB),
    FileItem(path=[".pad", str(2 * KiB)], length=2 * KiB, attr=b"p"),
    FileItem(path=["sup"], length=3 * KiB),
    FileItem(path=[".pad", str(13 * KiB)], length=13 * KiB, attr=b"p"),
]
inconsistent_padding = [
    FileItem(path=["inconsistent"], length=5 * KiB),
    FileItem(path=["hey"], length=14 * KiB),
    FileItem(path=[".pad", str(2 * KiB)], length=2 * KiB, attr=b"p"),
    FileItem(path=["sup"], length=3 * KiB),
    FileItem(path=[".pad", str(13 * KiB)], length=13 * KiB, attr=b"p"),
]
incorrect_padding = [
    FileItem(path=["incorrect"], length=16 * KiB),
    FileItem(path=["hey"], length=14 * KiB),
    FileItem(path=[".pad", str(7 * KiB)], length=7 * KiB, attr=b"p"),
    FileItem(path=["sup"], length=14 * KiB),
    FileItem(path=[".pad", str(7 * KiB)], length=7 * KiB, attr=b"p"),
]


padding_cases = (no_padding, correct_padding, inconsistent_padding, incorrect_padding)


@pytest.mark.parametrize(
    "mode,valid",
    [
        ("default", (True, True, True, True)),
        ("strict", (False, True, False, False)),
        ("forbid", (True, False, False, False)),
        ("ignore", (True, True, True, True)),
    ],
)
def test_padding_validation(mode, valid):
    """
    See description of padding modes in ValidationContext typed dict
    """
    for case, is_valid in zip(padding_cases, valid):
        if is_valid:
            InfoDictV1Base.model_validate(
                dict(name="test", files=case, piece_length=16 * KiB), context={"padding": mode}
            )
        else:
            with pytest.raises(ValidationError):
                InfoDictV1Base.model_validate(
                    dict(name="test", files=case, piece_length=16 * KiB), context={"padding": mode}
                )



import pytest

from torrent_models import testing
from torrent_models.torrent import Torrent

__all__ = [
    "libtorrent_pair",
]


@pytest.fixture(
    params=[
        pytest.param("v1", id="v1", marks=pytest.mark.v1),
        pytest.param("v2", id="v2", marks=pytest.mark.v2),
        pytest.param("hybrid", id="hybrid", marks=pytest.mark.hybrid),
    ],
)
def libtorrent_pair(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
) -> tuple[dict, Torrent]:
    """
    A pair of basic torrents generated with libtorrent and our methods
    """
    tmp_path = tmp_path_factory.mktemp(request.param)
    version = request.param
    piece_length = 32 * (2**10)
    paths = testing.make_files(base=tmp_path, piece_length=piece_length)
    torrent = testing.default_torrent(
        paths=paths, base=tmp_path, piece_length=piece_length, version=version
    )
    libtorrent = testing.default_libtorrent(
        paths=paths, base=tmp_path, piece_length=piece_length, version=version
    )

    return libtorrent, torrent



import pytest

__all__ = ["version"]


@pytest.fixture(
    params=[
        pytest.param("v1", id="v1", marks=pytest.mark.v1),
        pytest.param("v2", id="v2", marks=pytest.mark.v2),
        pytest.param("hybrid", id="hybrid", marks=pytest.mark.hybrid),
    ],
)
def version(request: pytest.FixtureRequest) -> str:
    return request.param



import pytest
from pydantic import TypeAdapter

from torrent_models import KiB
from torrent_models.info import InfoDictHybrid, InfodictUnionType, InfoDictV1, InfoDictV2
from torrent_models.types import str_keys

v1_infodict = {
    b"name": b"sup.exe",
    b"pieces": b"0" * 20,
    b"piece length": 16 * KiB,
    b"length": 16 * KiB,
}
v2_infodict = {
    b"name": b"sup.exe",
    b"meta version": 2,
    b"piece length": 16 * KiB,
    b"file tree": {b"sup.exe": {b"": {b"length": 16 * KiB, b"pieces root": b"0" * 32}}},
}
hybrid_infodict = {**v1_infodict, **v2_infodict}
infodicts = {"v1": v1_infodict, "v2": v2_infodict, "hybrid": hybrid_infodict}
models = {"v1": InfoDictV1, "v2": InfoDictV2, "hybrid": InfoDictHybrid}


@pytest.mark.parametrize("version", ("v1", "v2", "hybrid"))
@pytest.mark.parametrize("mode", ("bytes", "str", "model"))
def test_infodict_union_discriminator(version, mode):
    info = infodicts[version].copy()
    if mode == "str":
        info = str_keys(info)
    elif mode == "model":
        info = models[version].model_validate(info)

    adapter = TypeAdapter(InfodictUnionType)
    v = adapter.validate_python(info)
    assert isinstance(v, models[version])



from torrent_models.types.common import (
    AbsPath,
    FileName,
    FilePart,
    GenericFileItem,
    RelPath,
    SHA1Hash,
    SHA256Hash,
    TorrentVersion,
    TrackerFields,
)
from torrent_models.types.serdes import (
    EXCLUDE_STRINGIFY,
    ByteStr,
    ByteUrl,
    ListOrValue,
    UnixDatetime,
    str_keys,
    str_keys_list,
)
from torrent_models.types.v1 import FileItem, Pieces, V1PieceLength
from torrent_models.types.v2 import (
    FileTree,
    FileTreeItem,
    FileTreeType,
    MerkleTree,
    MerkleTreeShape,
    PieceLayerItem,
    PieceLayers,
    PieceLayersType,
    V2PieceLength,
)

__all__ = [
    "EXCLUDE_STRINGIFY",
    "AbsPath",
    "ByteStr",
    "ByteUrl",
    "FileItem",
    "FileName",
    "FilePart",
    "FileTree",
    "FileTreeItem",
    "FileTreeType",
    "GenericFileItem",
    "ListOrValue",
    "MerkleTree",
    "MerkleTreeShape",
    "PieceLayerItem",
    "PieceLayers",
    "PieceLayersType",
    "Pieces",
    "RelPath",
    "SHA1Hash",
    "SHA256Hash",
    "TorrentVersion",
    "TrackerFields",
    "UnixDatetime",
    "V1PieceLength",
    "V2PieceLength",
    "str_keys",
    "str_keys_list",
]



import posixpath
from math import ceil
from pathlib import Path
from typing import Any, BinaryIO, Self, cast
from typing import Literal as L

import bencode_rs
import humanize
from pydantic import Field, model_validator
from rich import print
from rich.console import Group
from rich.pretty import Pretty
from rich.table import Table

from torrent_models.base import ConfiguredBase
from torrent_models.info import (
    InfoDictHybrid,
    InfodictUnionType,
    InfoDictV1,
    InfoDictV1Base,
    InfoDictV2,
    InfoDictV2Base,
)
from torrent_models.types import (
    ByteStr,
    FileItem,
    FileTreeItem,
    GenericFileItem,
    ListOrValue,
    PieceLayersType,
    TorrentVersion,
    UnixDatetime,
    str_keys,
)
from torrent_models.types.v1 import FileItemRange, V1PieceRange
from torrent_models.types.v2 import FileTree, V2PieceRange


class TorrentBase(ConfiguredBase):
    announce: ByteStr | None = None
    announce_list: list[list[ByteStr]] | None = Field(default=None, alias="announce-list")
    comment: ByteStr | None = None
    created_by: ByteStr | None = Field(None, alias="created by")
    creation_date: UnixDatetime | None = Field(default=None, alias="creation date")
    info: InfodictUnionType
    piece_layers: PieceLayersType | None = Field(None, alias="piece layers")
    url_list: ListOrValue[ByteStr] | None = Field(
        None, alias="url-list", description="List of webseeds"
    )

    _flat_files: dict[str, FileTreeItem] | None = None
    _files: list[GenericFileItem] | None = None

    @property
    def webseeds(self) -> list[str] | None:
        """alias to url_list"""
        return self.url_list

    @classmethod
    def read_stream(cls, stream: BinaryIO, context: dict | None = None) -> Self:
        tdata = stream.read()
        tdict = bencode_rs.bdecode(tdata)
        return cls.from_decoded(decoded=tdict, context=context)

    @classmethod
    def read(cls, path: Path | str, context: dict | None = None) -> Self:
        with open(path, "rb") as tfile:
            torrent = cls.read_stream(tfile, context=context)
        return torrent

    @classmethod
    def from_decoded(
        cls, decoded: dict[str | bytes, Any], context: dict | None = None, **data: Any
    ) -> Self:
        """Create from bdecoded dict"""
        if decoded is not None:
            # we fix these incompatible types in str_keys
            decoded.update(data)  # type: ignore
            data = decoded  # type: ignore

        if any([isinstance(k, bytes) for k in data]):
            data = str_keys(data)  # type: ignore

        if context is None:
            context = {}

        return cls.model_validate(data, context=context)

    @property
    def torrent_version(self) -> TorrentVersion:
        if isinstance(self.info, InfoDictV1Base) and not isinstance(self.info, InfoDictV2Base):
            return TorrentVersion.v1
        elif isinstance(self.info, InfoDictV2Base) and not isinstance(self.info, InfoDictV1Base):
            return TorrentVersion.v2
        else:
            return TorrentVersion.hybrid

    @property
    def v1_infohash(self) -> str | None:
        """hex-encoded SHA1 of the infodict"""
        return self.info.v1_infohash

    @property
    def v2_infohash(self) -> str | None:
        """hex-encoded SHA256 of the infodict"""
        return self.info.v2_infohash

    @property
    def n_files(self) -> int:
        """
        Total number of files described by the torrent, excluding padfiles
        """

        if self.torrent_version in (TorrentVersion.v1, TorrentVersion.hybrid):
            self.info = cast(InfoDictV1 | InfoDictHybrid, self.info)
            if self.info.files is None:
                return 1
            return len([f for f in self.info.files if f.attr not in (b"p", "p")])
        else:
            self.info = cast(InfoDictV2, self.info)
            tree = FileTree.flatten_tree(self.info.file_tree)
            return len(tree)

    @property
    def total_size(self) -> int:
        """
        Total size of the torrent, excluding padfiles, in bytes
        """
        if self.torrent_version in (TorrentVersion.v1, TorrentVersion.hybrid):
            self.info = cast(InfoDictV1 | InfoDictHybrid, self.info)
            if self.info.files is None:
                self.info.length = cast(int, self.info.length)
                return self.info.length
            return sum([f.length for f in self.info.files if f.attr not in (b"p", "p")])
        else:
            self.info = cast(InfoDictV2, self.info)
            tree = FileTree.flatten_tree(self.info.file_tree)
            return sum([t["length"] for t in tree.values()])

    @property
    def flat_files(self) -> dict[str, FileTreeItem] | None:
        """A flattened version of the v2 file tree"""
        if self._flat_files is None and self.torrent_version != TorrentVersion.v1:
            self.info = cast(InfoDictV2, self.info)
            self._flat_files = FileTree.flatten_tree(self.info.file_tree)
        return self._flat_files

    @property
    def files(self) -> list[GenericFileItem]:
        """
        Common access to file information from both v1 and v2 torrents
        """
        if self._files is None:
            # v1 and v2 reps already confirmed to be equivalent during validation
            files = []
            if self.torrent_version in (TorrentVersion.v1, TorrentVersion.hybrid):
                self.info = cast(InfoDictV1 | InfoDictHybrid, self.info)
                if self.info.files is None:
                    v1_files = [FileItem(length=self.info.length, path=[self.info.name])]
                else:
                    v1_files = self.info.files
                for f in v1_files:
                    if f.is_padfile:
                        continue
                    v1_repr = f.model_dump()
                    v1_repr["path"] = posixpath.join(*v1_repr["path"])
                    if isinstance(v1_repr["path"], bytes):
                        v1_repr["path"] = v1_repr["path"].decode("utf-8")
                    if self.torrent_version == TorrentVersion.hybrid:
                        v2_repr = self.flat_files[v1_repr["path"]]  # type: ignore
                    else:
                        v2_repr = {}
                    files.append(GenericFileItem(**{**v1_repr, **v2_repr}))
            else:
                files = [GenericFileItem(path=k, **v) for k, v in self.flat_files.items()]  # type: ignore
            self._files = files
        return self._files

    @property
    def flat_trackers(self) -> list[list[str]]:
        trackers = []
        if self.announce:
            trackers.append([self.announce])
        if self.announce_list:
            trackers.extend(self.announce_list)
        return trackers

    def model_dump_torrent(self, mode: L["str", "binary"] = "str", **kwargs: Any) -> dict:
        """
        Dump the model into a dictionary that can be bencoded into a torrent

        Args:
            mode ("str", "binary"): ``str`` returns as a 'python' version of the torrent,
                with string keys and serializers applied.
                ``binary`` roundtrips to and from bencoding.
            kwargs: forwarded to :meth:`pydantic.BaseModel.model_dump`
        """
        dumped = self.model_dump(exclude_none=True, by_alias=True, **kwargs)
        if mode == "binary":
            dumped = bencode_rs.bdecode(bencode_rs.bencode(dumped))
        return dumped

    def pprint(self, verbose: int = 0) -> None:
        """
        Pretty print the torrent.

        See :func:`.pprint`
        """
        pprint(self, verbose=verbose)


class Torrent(TorrentBase):
    """
    A valid torrent file, including hashes.
    """

    @property
    def file_size(self) -> int:
        """Size of the generated torrent file, in bytes"""
        return len(self.bencode())

    def bencode(self) -> bytes:
        dumped = self.model_dump_torrent(mode="str")
        return bencode_rs.bencode(dumped)

    def write(self, path: Path) -> None:
        """Write the torrent to disk"""
        with open(path, "wb") as f:
            f.write(self.bencode())

    def v1_piece_range(self, piece_idx: int) -> V1PieceRange:
        """Get a v1 piece range from the piece index"""
        assert self.torrent_version in (
            TorrentVersion.v1,
            TorrentVersion.hybrid,
        ), "Cannot get v1 piece ranges for v2-only torrents"
        self.info = cast(InfoDictV1 | InfoDictHybrid, self.info)
        if piece_idx >= len(self.info.pieces):
            raise IndexError(
                f"Cannot get piece index {piece_idx} for torrent with "
                f"{len(self.info.pieces)} pieces"
            )

        start_range = piece_idx * self.info.piece_length
        end_range = (piece_idx + 1) * self.info.piece_length

        if self.info.files is None:
            self.info.length = cast(int, self.info.length)
            # single file torrent
            return V1PieceRange(
                piece_idx=piece_idx,
                piece_hash=self.info.pieces[piece_idx],
                ranges=[
                    FileItemRange(
                        path=[self.info.name],
                        length=self.info.length,
                        range_start=start_range,
                        range_end=min(self.info.length, end_range),
                    )
                ],
            )

        size_idx = 0
        file_idx = 0
        found_len = 0
        ranges = []
        # first, find file where range starts
        # could probably be combined with the second step,
        # but just getting this working before worrying about aesthetics
        for i, file in enumerate(self.info.files):
            if file.length + size_idx > start_range:
                # range starts in this file
                # create the range from the first file
                file_range_start = start_range % size_idx if size_idx > 0 else start_range
                file_range_end = min(file.length, file_range_start + self.info.piece_length)
                found_len += file_range_end - file_range_start
                ranges.append(
                    FileItemRange(
                        path=file.path,
                        attr=file.attr,
                        length=file.length,
                        range_start=file_range_start,
                        range_end=file_range_end,
                    )
                )

                # index additional files starting at the next file
                file_idx = i + 1
                break
            else:
                size_idx += file.length

        # then, iterate through files until the range or files are exhausted
        while found_len < self.info.piece_length and file_idx < len(self.info.files):
            file = self.info.files[file_idx]
            file_range_start = 0
            file_range_end = min(file.length, self.info.piece_length - found_len)

            ranges.append(
                FileItemRange(
                    path=file.path,
                    attr=file.attr,
                    length=file.length,
                    range_start=file_range_start,
                    range_end=file_range_end,
                )
            )
            found_len += file_range_end - file_range_start
            file_idx += 1
        return V1PieceRange(
            piece_idx=piece_idx, ranges=ranges, piece_hash=self.info.pieces[piece_idx]
        )

    def v2_piece_range(self, file: str, piece_idx: int = 0) -> V2PieceRange:
        """
        Get a v2 piece range from a file path and optional piece index.

        If `piece_idx` is not provided (default to 0)...

        - If the file is larger than the piece length, gets the 0th piece.
        - If the file is smaller than the piece length,
          the range corresponds to the whole file, the hash is the root hash,
          and piece_idx is ignored.
        """
        assert self.torrent_version in (
            TorrentVersion.v2,
            TorrentVersion.hybrid,
        ), "Cannot get v2 piece ranges from a v1-only torrent"

        # satisfy mypy...
        self.info = cast(InfoDictV2 | InfoDictHybrid, self.info)
        flat_files = self.flat_files
        flat_files = cast(dict[str, FileTreeItem], flat_files)
        self.piece_layers = cast(PieceLayersType, self.piece_layers)

        if file not in flat_files:
            raise ValueError(f"file {file} not found in torrent!")

        root = flat_files[file]["pieces root"]

        if root not in self.piece_layers:
            # smaller then piece_length, piece range is whole file
            return V2PieceRange(
                piece_idx=0,
                path=file,
                range_start=0,
                range_end=flat_files[file]["length"],
                piece_length=self.info.piece_length,
                file_size=flat_files[file]["length"],
                root_hash=root,
            )
        else:
            if piece_idx >= len(self.piece_layers[root]):
                raise IndexError(
                    f"piece index {piece_idx} is out of range for file with "
                    f"{len(self.piece_layers[root])} pieces"
                )
            return V2PieceRange(
                piece_idx=piece_idx,
                path=file,
                range_start=piece_idx * self.info.piece_length,
                range_end=min(flat_files[file]["length"], (piece_idx + 1) * self.info.piece_length),
                piece_length=self.info.piece_length,
                file_size=flat_files[file]["length"],
                piece_hash=self.piece_layers[root][piece_idx],
                root_hash=root,
            )

    @model_validator(mode="after")
    def piece_layers_if_v2(self) -> Self:
        """If we are a v2 or hybrid torrent, we should have piece layers"""
        if self.torrent_version in (TorrentVersion.v2, TorrentVersion.hybrid):
            assert self.piece_layers is not None, "Hybrid and v2 torrents must have piece layers"
        return self

    @model_validator(mode="after")
    def pieces_layers_correct(self) -> Self:
        """
        All files with a length longer than the piece length should be in piece layers,
        Piece layers should have the correct number of hashes
        """
        if self.torrent_version == TorrentVersion.v1:
            return self
        self.piece_layers = cast(PieceLayersType, self.piece_layers)
        self.info = cast(InfoDictV2 | InfoDictHybrid, self.info)
        for path, file_info in self.info.flat_tree.items():
            if file_info["length"] > self.info.piece_length:
                assert file_info["pieces root"] in self.piece_layers, (
                    f"file {path} does not have a matching piece root in the piece layers dict. "
                    f"Expected to find: {file_info['pieces root']}"  # type: ignore
                )
                expected_pieces = ceil(file_info["length"] / self.info.piece_length)
                assert len(self.piece_layers[file_info["pieces root"]]) == expected_pieces, (
                    f"File {path} does not have the correct number of piece hashes. "
                    f"Expected {expected_pieces} hashes from file length {file_info['length']} "
                    f"and piece length {self.info.piece_length}. "
                    f"Got {len(self.piece_layers[file_info['pieces root']])}"
                )
        return self


def pprint(t: TorrentBase, verbose: int = 0) -> None:
    """
    Print the contents of a torrent file.

    By default, prints only the top-level metadata in a way that should always be
    smaller than one screen.

    Increase verbosity to show more of the torrent.

    Hashes are printed as hexadecimal numbers and split into individual pieces,
    but they are properly encoded in the torrent.

    Args:
        t (:class:`.Torrent`): The torrent to print.
        verbose (int): Level of detail to print.

            * ``1`` show files in separate table
            * ``2`` show truncated v1 piece hashes
            * ``3`` show everything as-is

    """
    # summary stats
    summary = {
        "# Files": humanize.number.intcomma(t.n_files),
        "Total Size": humanize.naturalsize(t.total_size, binary=True),
        "Piece Size": humanize.naturalsize(t.info.piece_length, binary=True),
    }
    if hasattr(t, "file_size"):
        summary["Torrent Size"] = humanize.naturalsize(t.file_size, binary=True)

    v1_infohash = t.v1_infohash
    v2_infohash = t.v2_infohash
    if v1_infohash:
        summary["V1 Infohash"] = v1_infohash
    if v2_infohash:
        summary["V2 Infohash"] = v2_infohash
    table = Table(title=t.info.name, show_header=False)
    table.add_column("", justify="left", style="magenta bold", no_wrap=True)
    table.add_column("")
    for k, v in summary.items():
        table.add_row(k, v)

    exclude = {}
    context = {"mode": "print", "hash_truncate": True}
    file_table = None
    if verbose <= 1:
        exclude = {"info": {"pieces", "file tree", "file_tree", "files"}, "piece_layers": True}
    elif verbose <= 2:
        exclude = {"info": {"file tree", "file_tree", "files"}, "piece_layers": True}
    else:
        context["hash_truncate"] = False

    # make file table
    if 1 <= verbose <= 2:
        file_table = Table(title="Files")
        file_table.add_column("Path", no_wrap=True)
        file_table.add_column("Size")

        if t.torrent_version == TorrentVersion.v1:
            t.info = cast(InfoDictV1, t.info)
            tfiles = (
                t.info.files
                if t.info.files is not None
                else [FileItem(path=t.info.name, length=t.info.length)]
            )

            files = [
                ("/".join(f.path), humanize.naturalsize(f.length, binary=True), "")
                for f in tfiles
                if f.attr not in (b"p", "p")
            ]
        else:
            t.info = cast(InfoDictV2 | InfoDictHybrid, t.info)
            file_table.add_column("Hash")
            tree = t.flat_files
            assert tree is not None
            files = [
                (
                    str(k),
                    humanize.naturalsize(v["length"], binary=True),
                    v["pieces root"].hex()[0:8],
                )
                for k, v in tree.items()
            ]

        for f in files:
            file_table.add_row(*f)

    dumped = t.model_dump(
        by_alias=True, exclude=exclude, exclude_none=True, context=context  # type: ignore
    )

    if verbose < 1 or verbose > 2:
        group = Group(
            table,
            Pretty(dumped),
        )
    elif verbose <= 2:
        assert file_table is not None
        group = Group(table, file_table, Pretty(dumped))

    print(group)



"""
Routines that interaact with torrents using libtorrent

https://www.rasterbar.com/products/libtorrent/index.html

Libtorrent should be imported nowhere else in the library,
because libtorrent is optional.

When using functions from this module elsewhere in the library,
import them within the scope of the function or method,
not the module.
"""

from pathlib import Path
from typing import TYPE_CHECKING, cast

import bencode_rs
from tqdm import tqdm

from torrent_models.compat import get_size
from torrent_models.types.common import TorrentVersion

if TYPE_CHECKING:

    from torrent_models import TorrentCreate

try:
    import libtorrent
except ImportError as e:
    raise ImportError(
        "libtorrent is not installed."
        "install it with pip or with the optional dependency group torrent-models[libtorrent]"
    ) from e


def create_from_model(
    torrent: "TorrentCreate",
    version: TorrentVersion | str = TorrentVersion.hybrid,
    output: Path | None = None,
    bencode: bool = False,
    progress: bool = False,
) -> dict | bytes:
    """
    Create a .torrent file using libtorrent

    .. admonition:: Incomplete!
        :class: warning

        This function is not guaranteed to dump all fields set in your model.
        Libtorrent is an imperative function creator and can't set arbitrary metadata.
        This is primarily intended for ensuring backwards compatibility and testing

    Args:
        torrent (:class:`.TorrentCreate`): Torrent creation object
        version (:class:`.TorrentVersion`): Torrent version to create - strings or enum values
        output (Path | None): If present, write torrent to file
        bencode (bool): If True, bencode before returning
        progress (bool): If True, show progress bar while hashing
    """
    if isinstance(version, str):
        version = TorrentVersion.__members__[version]

    fs = libtorrent.file_storage()
    for path in torrent.get_paths():
        fs.add_file(str(path), get_size(torrent.path_root / path))

    flags = 0
    if version == TorrentVersion.v1:
        flags = libtorrent.create_torrent.v1_only
    elif version == TorrentVersion.v2:
        flags = libtorrent.create_torrent.v2_only

    torrent.piece_length = cast(int, torrent.piece_length)
    created = libtorrent.create_torrent(fs, piece_size=torrent.piece_length, flags=flags)

    trackers = torrent.get_trackers()
    flat_trackers = [trackers.get("announce", None)]

    for tier in trackers.get("announce-list", []):
        flat_trackers.extend(tier)

    flat_trackers = list(dict.fromkeys([str(t) for t in flat_trackers if t is not None]))

    for tier, tracker in enumerate(flat_trackers):
        created.add_tracker(tracker, tier)

    if torrent.webseeds:
        for webseed in torrent.webseeds:
            created.add_url_seed(str(webseed))

    if torrent.similar:
        for s in torrent.similar:
            created.add_similar_torrent(libtorrent.sha1_hash(s))

    if torrent.comment:
        created.set_comment(torrent.comment)

    if torrent.created_by is not None:
        created.set_creator(torrent.created_by)

    _pbar = None
    if progress:
        _pbar = tqdm(desc="hashing pieces...", total=created.num_pieces())

        def _pbar_callback(piece_index: int) -> None:
            _pbar.update()

        libtorrent.set_piece_hashes(created, str(torrent.path_root.resolve()), _pbar_callback)
        _pbar.close()
    else:
        libtorrent.set_piece_hashes(created, str(torrent.path_root.resolve()))

    ret = created.generate()
    if bencode or output:
        bencoded = bencode_rs.bencode(ret)
        if output:
            with open(output, "wb") as f:
                f.write(bencoded)
        if bencode:
            return bencoded

    return ret



from functools import cached_property
from itertools import count
from multiprocessing.pool import AsyncResult
from multiprocessing.pool import Pool as PoolType
from pathlib import Path
from typing import TYPE_CHECKING, cast, overload

from pydantic import PrivateAttr, ValidationInfo, field_validator

from torrent_models.hashing.base import Chunk, Hash, HasherBase
from torrent_models.types.v1 import V1PieceLength

if TYPE_CHECKING:
    from multiprocessing.pool import AsyncResult


class V1Hasher(HasherBase):
    piece_length: V1PieceLength
    _buffer: bytearray = PrivateAttr(default_factory=bytearray)
    _v1_counter: count = PrivateAttr(default_factory=count)
    _last_path: Path | None = None

    @field_validator("read_size", mode="before")
    def read_size_is_piece_length(cls, value: int | None, info: ValidationInfo) -> int:
        """If read_size not passed, make it piece_length"""
        if value is None:
            value = info.data["piece_length"]
        return value

    @field_validator("paths", mode="after")
    def sort_paths(cls, value: list[Path]) -> list[Path]:
        """
        v1 torrents have arbitrary file sorting,
        but we mimick libtorrent/qbittorrent's sort order for consistency's sake
        """
        value = sort_v1(value)
        return value

    @overload
    def update(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]: ...

    @overload
    def update(self, chunk: Chunk, pool: None) -> list[Hash]: ...

    def update(self, chunk: Chunk, pool: PoolType | None = None) -> list[AsyncResult] | list[Hash]:
        return self._update_v1(chunk, pool)

    @cached_property
    def total_hashes(self) -> int:
        return self._v1_total_hashes()

    @overload
    def _update_v1(self, chunk: Chunk, pool: PoolType) -> list[AsyncResult]: ...

    @overload
    def _update_v1(self, chunk: Chunk, pool: None) -> list[Hash]: ...

    def _update_v1(self, chunk: Chunk, pool: PoolType | None) -> list[AsyncResult] | list[Hash]:
        # gather v1 pieces until we have blocks_per_piece or reach the end of a file
        # have to annotate as a mixed list here and recast at the end
        # because mypy can't follow the overloads
        res: list[AsyncResult | Hash] = []
        if self._last_path is not None and chunk.path != self._last_path and len(self._buffer) > 0:
            res.extend(self._on_file_end(pool=pool))
        self._last_path = chunk.path

        if len(chunk.chunk) == self.piece_length and not self._buffer:
            # shortcut for when our read_length is piece size -
            # don't copy to buffer if we don't have to
            chunk.idx = next(self._v1_counter)
            if pool:
                res.append(pool.apply_async(self._hash_v1, (chunk, self.path_root)))
            else:
                res.append(self._hash_v1(chunk, self.path_root))
        else:
            # handle file ends, read sizes that are larger/smaller than piece size.
            self._buffer.extend(chunk.chunk)
            while len(self._buffer) >= self.piece_length:
                piece = self._buffer[: self.piece_length]
                del self._buffer[: self.piece_length]
                piece_chunk = Chunk.model_construct(
                    idx=next(self._v1_counter), path=chunk.path, chunk=piece
                )
                if pool:
                    res.append(pool.apply_async(self._hash_v1, (piece_chunk, self.path_root)))
                else:
                    res.append(self._hash_v1(piece_chunk, self.path_root))

        res = cast(list[AsyncResult] | list[Hash], res)
        return res

    @overload
    def _on_file_end(self, pool: PoolType) -> list[AsyncResult]: ...

    @overload
    def _on_file_end(self, pool: None) -> list[Hash]: ...

    def _on_file_end(self, pool: PoolType | None) -> list[AsyncResult] | list[Hash]:
        """Allow hybrid hasher to add padding at file boundaries and submit piece"""
        return []

    @overload
    def _after_read(self, pool: PoolType) -> list[AsyncResult]: ...

    @overload
    def _after_read(self, pool: None) -> list[Hash]: ...

    def _after_read(self, pool: PoolType | None) -> list[AsyncResult] | list[Hash]:
        """Submit the final incomplete piece"""
        if len(self._buffer) > 0:
            self._last_path = cast(Path, self._last_path)
            chunk = Chunk.model_construct(
                idx=next(self._v1_counter), path=self._last_path, chunk=self._buffer
            )
            if pool:
                return [pool.apply_async(self._hash_v1, args=(chunk, self.path_root))]
            else:
                return [self._hash_v1(chunk, self.path_root)]
        else:
            return []


def sort_v1(paths: list[Path]) -> list[Path]:
    """
    v1 sorts top-level files first, then within that division alphabetically
    https://github.com/alanmcgovern/monotorrent/issues/563
    """
    return sorted(paths, key=lambda path: (len(path.parts) != 1, str(path).lower()))



from importlib.metadata import version

DEFAULT_TORRENT_CREATOR = f"torrent-models ({version('torrent-models')})"
EXCLUDE_FILES = (".DS_Store", "Thumbs.db")

KiB = 2**10
MiB = 2**20
GiB = 2**30
TiB = 2**40


class size:
    """Convenience accessor/importer for sizes"""

    KiB = KiB
    MiB = MiB
    GiB = GiB
    TiB = TiB


BLOCK_SIZE = 16 * KiB



import posixpath
import random
import string
from math import ceil
from pathlib import Path

import pytest

from torrent_models import KiB, TorrentCreate, TorrentVersion
from torrent_models.const import EXCLUDE_FILES

SIZES = [10 * KiB, 20 * KiB, 32 * KiB, 40 * KiB, 100 * KiB]


@pytest.fixture(params=SIZES)
def file_size(request: pytest.FixtureRequest, tmp_path: Path) -> int:
    """Create a set of files that are smaller than, equal to, and larger than a 32 KiB piece size"""

    size = request.param
    for name in string.ascii_letters[0:10]:
        with open(tmp_path / name, "wb") as f:
            f.write(random.randbytes(size))
    return size


@pytest.mark.parametrize("version", [TorrentVersion.v1, TorrentVersion.hybrid])
def test_v1_piece_range(file_size: int, version: TorrentVersion, tmp_path: Path):
    """
    We can get piece ranges from v1 torrents and validate data against them
    """
    files = [p for p in tmp_path.iterdir() if p.name not in EXCLUDE_FILES]
    assert len(files) == 10
    assert all([(tmp_path / p).stat().st_size == file_size for p in files])

    create = TorrentCreate(paths=files, path_root=tmp_path, piece_length=32 * KiB)
    torrent = create.generate(version=version)
    seen_files = set()
    for i, piece in enumerate(torrent.info.pieces):
        range = torrent.v1_piece_range(i)
        assert range.piece_hash == piece
        data = []
        for file in range.ranges:
            if file.is_padfile:
                data.append(bytes(file.range_end - file.range_start))
            else:
                path = posixpath.join(*file.path)
                seen_files.add(path)
                with open(tmp_path / path, "rb") as f:
                    f.seek(file.range_start)
                    data.append(f.read(file.range_end - file.range_start))

        assert range.validate_data(data)

        # we reject random data in the right shape
        fake_data = [random.randbytes(len(d)) for d in data]
        assert not range.validate_data(fake_data)

    assert seen_files == {letter for letter in string.ascii_letters[0:10]}


@pytest.mark.parametrize("version", [TorrentVersion.v2, TorrentVersion.hybrid])
def test_v2_piece_range(file_size: int, version: TorrentVersion, tmp_path: Path):
    """
    We can get piece ranges from v2 torrents and validate data against them
    """
    files = [p for p in tmp_path.iterdir() if p.name not in EXCLUDE_FILES]
    assert len(files) == 10
    assert all([(tmp_path / p).stat().st_size == file_size for p in files])

    create = TorrentCreate(paths=files, path_root=tmp_path, piece_length=32 * KiB)
    torrent = create.generate(version=version)
    assert set(torrent.flat_files.keys()) == {letter for letter in string.ascii_letters[0:10]}
    for path, file_info in torrent.flat_files.items():
        root = file_info["pieces root"]
        n_pieces = 1 if root not in torrent.piece_layers else len(torrent.piece_layers[root])
        assert n_pieces == ceil(file_size / (32 * KiB))
        for piece_idx in range(n_pieces):
            piece_range = torrent.v2_piece_range(path, piece_idx)
            assert piece_range.range_start == piece_idx * 32 * KiB

            with open(tmp_path / path, "rb") as f:
                f.seek(piece_range.range_start)
                data = f.read(piece_range.range_end - piece_range.range_start)
            data = [data[i : i + (16 * KiB)] for i in range(0, len(data), 16 * KiB)]
            assert piece_range.validate_data(data)

            # reject random data in the right shape
            data = [random.randbytes(len(d)) for d in data]
            assert not piece_range.validate_data(data)
